import kfp
import kfp.components._components as comp
from typing import Callable, List, Dict, Any, Optional, NamedTuple
from importlib.util import module_from_spec, spec_from_loader
from types import ModuleType
from kfp.components import InputPath, OutputPath
from pathlib import Path
from datetime import datetime
import os.path
import sys
import json
import tempfile
import zipfile
import subprocess


class NotebookComponent:
    """Notebook Component is used to create a component by submitting a notebook and its dependencies to GCS then download, extract the archived zip file and run it at runtime.
    Notebook outputs are visualized as static HTML which could be retrieved at `Artifact` Tab or `Run output` Tab in the Kubeflow Pipeline UI.
    Parameters
    ----------
    notebook: str
        Path to notebook
    dependencies: str
        Path to dependencies
    gs_path: str
        GCS path to store archived zip file
    output_component_file: Optional[str]
        Export yaml format to file
    base_image: Optional[str]
        Image to run notebook
    packages_to_install: Optional[List[str]]
        Additional packages to install at runtime
    params_schema: Optional[Dict[str, Any]]
        Schema for parameters that are defined at notebook cell with "parameters" tag
    remove_nb_inputs: bool
        True to exclude input / input prompt. Default True
    """
    def __init__(
        self, 
        notebook: str,
        dependencies: str,
        gs_path: str,
        output_component_file: Optional[str] = None,
        base_image: Optional[str] = None,
        packages_to_install: Optional[List[str]] = [],
        params_schema: Optional[Dict[str, Any]] = None,
        remove_nb_inputs:bool = True
    ):
        self.notebook = Path(notebook.strip())
        self.dependencies = Path(dependencies.strip()) if dependencies else ''
        self.gs_path = gs_path.rstrip('/')
        self.output_component_file = output_component_file
        self.base_image = base_image
        self.packages_to_install = ["jupyter", "papermill", "nbconvert", "google-cloud-storage"] + packages_to_install
        self.zipname = ''
        self.params_schema = params_schema
        self.remove_nb_inputs = remove_nb_inputs
        
        if not self.notebook.is_file():
            raise "Input Notebook is not exist"
        
        if not self.base_image:
            raise "base_image must be defined"
            
        if not self.gs_path:
            raise "gs_path must be defined"
            
                    
        
    def load_component(self, version: Optional[str]=None):
        """Submit notebook to GCS and create Kubeflow Pipeline component.
        Parameters
        ----------
        version: Optional[str]
            Defined version of Notebook archive zip file. If not defined, version would be "YYYYMMDD_HHMMSS"
        
        Return
        ------
        KF Component
        """
        # Set version for notebook archive
        version = version or datetime.now().strftime('%Y%m%d_%H%M%S')
        
        # Zip notebook archive
        self.zipname = f"{self.notebook.name.split('.')[0]}_{version}.zip"
        tmpdir = tempfile.mkdtemp()
        zip_path = Path(tmpdir) / self.zipname
        zip_obj = zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED)
        zip_obj.write(self.notebook)
        if self.dependencies:
            for file in self.dependencies.glob('**/*'):
                zip_obj.write(file)
        zip_obj.close()
        
        # Upload notebook archived zip file to GCS
        cmd = f'gsutil -m cp {zip_path} {self.gs_path}/{zip_path.name}'
        process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        process.communicate()
        
        if self.params_schema:
            func_params = []
            pm_params = []
            for param_name, type_ in self.params_schema.items():
                func_params.append(f"{param_name}: {type_}")
                pm_params.append(f"'{param_name}': {param_name}")
            func_params = ",".join(func_params)
            pm_params = "{" + f"{','.join(pm_params)}" + "}"
        else:
            func_params = ""
            pm_params = "{}"
            
        execute_func = f"""def NotebookComponent(
    {func_params}
) -> NamedTuple('StaticHTMLOutput', [('mlpipeline_ui_metadata', 'UI_metadata')]):
    '''A wrapper func to run notebook'''
    import zipfile
    import json
    import papermill as pm
    from pathlib import Path
    from google.cloud import storage
    from nbconvert import HTMLExporter
    from traitlets.config import Config
    import os

    client = storage.Client()
    gs_path = '{self.gs_path}'[len('gs://'):].split('/')
    bucket, blob = gs_path[0], '/'.join(gs_path[1:] + ['{zip_path.name}'])
    
    bucket = client.get_bucket(bucket)
    
    blob = bucket.blob(blob)
    blob.download_to_filename('{zip_path.name}')
    
    print('Downloaded: {zip_path.name}')
    
        
    with zipfile.ZipFile('./{zip_path.name}', 'r') as zip_ref:
        zip_ref.extractall('.')
    
    pm.execute_notebook(
        './{self.notebook.name}',
        './output.ipynb',
        parameters={pm_params}
    )
    
    c = Config()
    if {self.remove_nb_inputs} == True:
        c.HTMLExporter.exclude_input_prompt = True
        c.HTMLExporter.exclude_input = True
        
    htmlExporter = HTMLExporter(config=c)
    htmlExporter.template_name = 'classic'
    body, _ = htmlExporter.from_filename('output.ipynb')
                
    metadata = {{
        "version": 1,
        'outputs' : [
        {{
            'type' : 'web-app',
            'storage' : 'inline',
            'source' : body
        }}]
    }}
    
    from collections import namedtuple
    static_html_output = namedtuple('StaticHTMLOutput', ['mlpipeline_ui_metadata'])
    return static_html_output(json.dumps(metadata))"""
        
        _import_script = "from kfp.components import *\nfrom typing import NamedTuple\n"
        execute_func = _import_script + execute_func
                
        filename = tempfile.mktemp(suffix='.py')
        modname = os.path.splitext(os.path.basename(filename))[0]
        assert modname not in sys.modules
        loader = ShowSourceLoader(modname, execute_func)
        spec = spec_from_loader(modname, loader, origin=filename)
        module = module_from_spec(spec)
        code = compile(execute_func, mode='exec', filename=filename)
        exec(code, module.__dict__)
        sys.modules[modname] = module
        
        return kfp.components.create_component_from_func(
            func=module.NotebookComponent,
            output_component_file=self.output_component_file,
            base_image=self.base_image,
            packages_to_install=self.packages_to_install
        )
        
        

class ShowSourceLoader:
    def __init__(self, modname: str, source: str) -> None:
        self.modname = modname
        self.source = source

    def get_source(self, modname: str) -> str:
        if modname != self.modname:
            raise ImportError(modname)
        return self.source

        
    
if __name__ == "__main__":
    from kfp.components import InputPath
    help(NotebookComponent(
        notebook="./Untitled.ipynb",
        dependencies="./kfpxtend",
        gs_path="gs://vinid-data-science-rec-ecart-np/tmp",
        base_image='hihi',
        params_schema={'a': "int", 'bigA': "InputPath('CSV')", 'b': "'GCS Path'"}
    ).load_component())


import requests
import subprocess as sub

class CloudBearerAuth(requests.auth.AuthBase):
    def __init__(self):
        process = sub.Popen(['gcloud', 'auth', 'print-identity-token'], stdout=sub.PIPE)
        token = process.communicate()[0].decode(encoding='utf-8')
        self.token = token.strip()
    
    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r



from datetime import datetime
from functools import wraps
import pytz

def _exception_handler(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print(e)
            return None
    return _wrapper


def show_examples():
    mess = ['Give kfpxtend a \U0001F31F on github', 
            'Examples:', 
            'https://nbviewer.jupyter.org/github/vule24/kfpxtend/blob/master/examples/basic.ipynb',
            'https://nbviewer.jupyter.org/github/vule24/kfpxtend/blob/master/examples/advance.ipynb']
    _max = max(map(len, mess))
    _max = 40 if _max < 40 else _max + 20
    print('╔' + '═'*_max + '╗')
    for i,m in enumerate(mess):
        _left = (_max - len(m))//2 + 1
        print(' '*_left + m)
    print('╚' + '═'*_max + '╝')    
          
@_exception_handler
def get_pipeline_by_name(self, name):
    """Get pipeline using its name.
    Parameters
    ----------
    name: str
        Name of pipeline.
        
    Returns
    -------
    str
        ID of pipeline
    """
    next_page_token = ''
    while True:
        pipelines_info = self.pipelines.list_pipelines(page_token=next_page_token, page_size=self.page_size)
        next_page_token = pipelines_info.next_page_token
        for pipeline in pipelines_info.pipelines:
            if pipeline.name == name:
                return pipeline.id
        if next_page_token is None:
            break
    return ''

@_exception_handler
def get_experiment_by_name(self, name):
    """Get experiment using its name.
    Parameters
    ----------
    name: str
        Name of experiment.
        
    Returns
    -------
    str
        ID of experiment.
    """
    next_page_token = ''
    while True:
        exps_info = self.experiments.list_experiment(page_token=next_page_token,page_size=self.page_size)
        next_page_token = exps_info.next_page_token
        for exp in exps_info.experiments:
            if exp.name == name:
                return exp.id
        if next_page_token is None:
            break
    return ''

@_exception_handler
def get_run_by_name(self, name, level:str='easy'):
    """Get run using its name.
    Parameters
    ----------
    name: str
        Name of run.
    level: str
        If `hard`, retrieve run till the last page.
        If `easy`, find the latest runs in the first page with `page_size` elements.
        
    Returns
    -------
    List[str]
        List of ids.
    """
    run_ids = []
    if level == 'easy':
        for run in self.runs.list_runs(page_size=self.page_size).runs:
            if run.name == name:
                run_ids.append(run.id)
    elif level == 'hard':
        next_page_token = ''
        while True:
            runs_info = self.runs.list_runs(page_token=next_page_token, page_size=self.page_size)
            next_page_token = runs_info.next_page_token
            for run in runs_info.runs:
                if run.name == name:
                    run_ids.append(run.id)
            if next_page_token is None:
                break
    return run_ids

@_exception_handler
def get_runs_by_experiment(self, exp_name, level:str='easy'):
    """Get runs that belongs to an experiment.
    Parameters
    ----------
    exp_name: str
        Experiment name.
    level: str
        If `hard`, retrieve run till the last page even if the exp_name has been removed.
        If `normal`, retrieve run till the last page when exp_name still exists.
        If `easy`, find the latest runs in the first page with `page_size` elements.
        
    Returns
    -------
    Dict[str, str]
        Dictionary of runs in experiment.
    """
    run_name_ids = []
    if level == 'easy':
        for run in self.runs.list_runs(page_size=self.page_size).runs:
            if run.resource_references[0].name == exp_name:
                run_name_ids.append({'name': run.name, 'id': run.id})
    elif level == 'normal':
        exp_id = self.get_experiment_by_name(exp_name)
        if not exp_id:
            print(f'Experiment: {exp_name} not exists')
            return {}
        next_page_token = ''
        while True:
            runs_info = self.list_runs(page_token=next_page_token, page_size=self.page_size, experiment_id=exp_id)
            next_page_token = runs_info.next_page_token
            for run in runs_info.runs:
                run_name_ids.append({'name': run.name, 'id': run.id})
            if next_page_token is None:
                break
    elif level == 'hard':
        next_page_token = ''
        while True:
            runs_info = self.list_runs(page_token=next_page_token, page_size=self.page_size)
            next_page_token = runs_info.next_page_token
            for run in runs_info.runs:
                is_match_exp = False
                for resource in run.resource_references:
                    if resource.key.type=='EXPERIMENT':
                        is_match_exp=True if resource.name == exp_name else False
                        break
                if is_match_exp:
                    run_name_ids.append({'name': run.name, 'id': run.id})
            if next_page_token is None:
                break
    else:
        raise NotImplementedError
    return run_name_ids
        
@_exception_handler
def delete_pipeline_by_name(self, name):
    """Delete pipeline using its name.
    Parameters
    ----------
    name: str
        Name of pipeline to be deleted.
    """
    pipeline_id = self.get_pipeline_by_name(name)        
    self.pipelines.delete_pipeline(pipeline_id)
    print(f'Deleted pipeline\tName: {name}\tID: {pipeline_id}')
    
@_exception_handler
def delete_experiment_by_name(self, name):
    """Delete experiment using its name.
    Parameters
    ----------
    name: str
        Name of experiment to be deleted.
    """
    exp_id = self.get_experiment_by_name(name)
    self.experiments.delete_experiment(exp_id)
    print(f'Deleted experiment\tName: {name}\tID: {exp_id}')
    
@_exception_handler
def delete_run_by_name(self, name, level:str='easy'):
    """Delete run using its name.
    Parameters
    ----------
    name: str
        Name of run to be deleted.
    level: str
        If `hard`, retrieve run till the last page.
        If `easy`, find the latest runs in the first page with `page_size` elements.
    """
    run_ids = self.get_run_by_name(name, level)
    print(f'Total: {len(run_ids)} runs')
    for i, run_id in enumerate(run_ids):
        self.runs.delete_run(run_id)
        print(f'Deleted run [{i}]:\tName: {name}\tID: {run_id}')
    
@_exception_handler
def delete_runs_by_experiment(self, exp_name, level:str='easy'):
    """Delete runs that belongs to an experiment.
    Parameters
    ----------
    exp_name: str
        Name of experiment.
    level: str
        If `hard`, retrieve run till the last page even if the exp_name has been removed.
        If `normal`, retrieve run till the last page when exp_name still exists.
        If `easy`, find the latest runs in the first page with `page_size` elements.
    """
    name_ids = self.get_runs_by_experiment(exp_name, level)
    print(f'Total: {len(name_ids)} runs')
    for i, run in enumerate(name_ids):
        self.runs.delete_run(run["id"])
        print(f'Deleted run [{i}]:\tName: {run["name"]}\tID: {run["id"]}')
        
@_exception_handler
def upload_pipeline_version(self, filepath, pipeline_name:str, version:str=None):
    """Upload a version of existing pipeline.
    Parameters
    ----------
    filepath: str
        Path to .yaml file.
    pipeline_name: str
        Name of existing pipeline.
    version: str, optional
        Version of new uploaded pipeline. Defaults to None to get version as timestamp.
    """
    version = '{name}_version_at_{time}'.format(
        name=pipeline_name,
        time=pytz.timezone("Asia/Ho_Chi_Minh").localize(datetime.now()).strftime('%Y-%m-%d_%H:%M:%S')
    ) if version is None else version
    
    try:
        self.upload_pipeline(filepath, pipeline_name)
        print('Uploaded new pipeline')
    except:
        pipelineid = self.get_pipeline_by_name(pipeline_name)
        self.pipeline_uploads.upload_pipeline_version(
            filepath,
            pipelineid=pipelineid,
            name=version)
        print(f'Uploaded version for {pipeline_name}')


def create_pipeline_workflow(self, parameter_defaults, pipeline, op_transformers=None, pipeline_conf=None):
    workflow = self.__create_pipeline_workflow(parameter_defaults, pipeline, op_transformers, pipeline_conf)
    if pipeline_conf.gcp_sa_name:
        workflow['spec']['serviceAccountName'] = pipeline_conf.gcp_sa_name
    return workflow


import copy
import kfp
import kfp.components._components as comp
from kfp.components._structures import ComponentReference

def load_component_from_gcs(uri):
    from google.cloud import storage
    uri = uri[len('gs://'):].split('/')
    client = storage.Client()
    blob = client.bucket(uri[0]).blob('/'.join(uri[1:]))
    return blob.download_as_string()


class GSComponentStore(kfp.components.ComponentStore):
    def __init__(self, local_search_paths=None, url_search_prefixes=None, gs_search_prefixes=None):
        self.gs_search_prefixes = gs_search_prefixes
        super().__init__(local_search_paths=local_search_paths, url_search_prefixes=url_search_prefixes)
    
    def _load_component_spec_in_component_ref(
        self,
        component_ref: ComponentReference,
    ) -> ComponentReference:
        """Takes component_ref, finds the component spec and returns component_ref with .spec set to the component spec.
        See ComponentStore.load_component for the details of the search logic.
        """
        if component_ref.spec:
            return component_ref

        component_ref = copy.copy(component_ref)
        if component_ref.url:
            component_ref.spec = comp._load_component_spec_from_url(url=component_ref.url, auth=self._auth)
            return component_ref

        name = component_ref.name
        if not name:
            raise TypeError("name is required")
        if name.startswith('/') or name.endswith('/'):
            raise ValueError('Component name should not start or end with slash: "{}"'.format(name))

        digest = component_ref.digest
        tag = component_ref.tag

        tried_locations = []

        if digest is not None and tag is not None:
            raise ValueError('Cannot specify both tag and digest')

        if digest is not None:
            path_suffix = name + '/' + self._digests_subpath + '/' + digest
        elif tag is not None:
            path_suffix = name + '/' + self._tags_subpath + '/' + tag
            #TODO: Handle symlinks in GIT URLs
        else:
            path_suffix = name + '/' + self._component_file_name

        #Trying URL prefixes
        for gs_search_prefix in self.gs_search_prefixes:
            uri = gs_search_prefix.rstrip('/') + '/' + path_suffix
            tried_locations.append(uri)
            try:
                if uri.startswith('gs://'):
                    component_content = load_component_from_gcs(uri)
                    component_ref.spec = comp._load_component_spec_from_yaml_or_zip_bytes(component_content)
                    return component_ref
            except:
                pass
        return super()._load_component_spec_in_component_ref(component_ref)
        
    def list(self):
        super().search('')
        from google.cloud import storage
        client = storage.Client()
        gcs_list = []
        for gs_search_prefix in self.gs_search_prefixes:
            uri = gs_search_prefix[len('gs://'):].split('/')
            store_blob = '/'.join(uri[1:])
            blobs = client.list_blobs(bucket_or_name=uri[0], prefix=store_blob, delimiter=None)

            for blob in blobs:
                if store_blob in blob.name and blob.name.endswith('component.yaml'):
                    gcs_list.append(['gs://' + '/'.join([uri[0], blob.name]), blob.name[len(store_blob):-len('/component.yaml')]])
        return gcs_list
    
    def search(self, name:str):
        super().search(name)
        for blob, component_name in self.list():
            if name.casefold() in blob.casefold():
                print(f"name: {component_name}  uri: {blob}")





from kfp import dsl
from pathlib import Path
from datetime import datetime
import zipfile
import subprocess


# TODO: Change ContainerOp to python_component

class NotebookOp(dsl.ContainerOp):
    """Represents an op implemented by a Notebook and dependencies.
    
    Args:
        notebook: path to notebook.
        dependencies: path to dependencies dir.
        gcs_bucket: Google Cloud Storage bucket.
        gcs_directory: GCS dir in gcs_bucket.
        version: tag version for each notebook run. If not define,
            version will be set to `%Y-%m-%d_%H:%M:%S` format.
        kwargs: all the arguments for ContainerOp.
        
    Note: 
        - This Op need to be initialized inside pipeline definition or inside a function
            that return an Op for triggering upload source to GCS whenever sources are updated.
        - NotebookOp use celltags: `params` to set parameters and `skip` to ignore cells, 
            others would be use as main source.
        - If arguments include list or dict, we can define them in string.
            Eg. ['--a', '"[1,2,3]"',
                 '--b', '\\"{"key1": 1, "key2": "value"}\\"']
        - Remember to save your Notebook before defining pipeline.
    Example: 
        >>> @dsl.pipeline(
        ...     name="test",
        ...     description="test"
        ... )
        ... def pipeline():
        ...     op = NotebookOp(
        ...         notebook='path/to/notebook.ipynb',
        ...         dependencies='path/to/dependencies',
        ...         gcs_bucket='bucket_name',
        ...         gcs_directory='path/to/gsc_directories',
        ...         version='test',
        ...         name='NotebookOpTest',
        ...         image='path/image:tag',
        ...         arguments=['--a', '"[1,2,3]"',
        ...                    '--b', '5']
        ...     )
    """
    def __init__(self,
                 notebook:str,
                 dependencies:str,
                 gcs_bucket:str,
                 gcs_directory:str,
                 version:str=None,
                 **kwargs):
        self.notebook = Path(notebook.strip())
        self.dependencies = Path(dependencies.strip('/')) if dependencies else None
        self.gcs_bucket = gcs_bucket.strip('/')
        self.gcs_directory = gcs_directory.strip('/')
        self.zipname = ''
        self.version = datetime.now().strftime("%Y-%m-%d_%H:%M:%S") if not version else version

        assert self.notebook.suffix == '.ipynb'
        if not notebook:
            raise ValueError("You need to provide a notebook.")
            
        if 'name' not in kwargs:
            raise TypeError("You need to provide a name for the operation.")
        elif not kwargs.get('name'):
            raise ValueError("You need to provide a name for the operation.")
        self.zipname = f"{kwargs['name']}_{self.version}.zip"

        if 'image' not in kwargs:
            raise ValueError("You need to provide an image.")
            
        kwargs['command'] = ['sh', '-c']
        kwargs['arguments'] = ' '.join(
            [(f'echo "SETUP:\n------" && '
              f'gsutil -m cp -r gs://{self.gcs_bucket}/{self.gcs_directory}/{self.zipname} . && '
              f'python3 -m zipfile -e {self.zipname} . && '
              f'echo "\nSOURCE:\n-------" && '
              f'python3 nb_convert.py && '
              f'echo "\nLOG:\n----" && '
              f'python3 {self.notebook.with_suffix(".py").name}'), 
             ' '.join(kwargs.get('arguments', []))]
        )
        self.upload_all()
        super().__init__(**kwargs)
        
    
    def get_converter_source(self):
        import inspect, os
        converter_src = list(filter('\n'.__ne__, inspect.getsourcelines(self.convert)[0][1:]))
        common = os.path.commonprefix(converter_src)
        ret = [line[len(common):] for line in converter_src]
        for i,line in enumerate(ret):
            ret[i] = line.replace("nb_path = Path(f'{notebook}')", f"nb_path = Path(f'{self.notebook.name}')")
        return r'{}'.format(''.join(ret))
        
    def convert(self, notebook):
        import json
        import itertools
        from pathlib import Path

        nb_path = Path(f'{notebook}')
        ipynb = json.loads(nb_path.read_text())
        cells = [cell for cell in ipynb['cells'] if cell['cell_type'] == 'code' and 'skip' not in cell['metadata'].get('tags', [])]

        params = list(itertools.chain.from_iterable([cell['source'] for cell in cells if 'params' in cell['metadata'].get('tags', [])]))
        source = list(itertools.chain.from_iterable([cell['source'] for cell in cells if 'params' not in cell['metadata'].get('tags', [])]))

        main_src = ('import argparse\n'
                    'import ast\n'
                    'import subprocess\n'
                    'parser = argparse.ArgumentParser()\n')

        params = list(map(lambda x: x.strip(), params))
        lvalues = []
        for expression in params:
            lvalue = expression.split('=')[0].strip()
            lvalues.append(lvalue)
        add_arguments = casts = ''
        for arg in lvalues:
            add_arguments += f"parser.add_argument('--{arg}', type=str, required=True)\n"
            casts += ("try:\n"
                      f"    {arg} = ast.literal_eval(args.{arg})\n"
                      "except:\n"
                      f"    {arg} = args.{arg}\n")

        main_src = ''.join([main_src, add_arguments, "args = parser.parse_args()\n", casts]) + '\n'.join(map(lambda x: x.strip('\n'), source))
        main_src = main_src.split('\n')
        for i, line in enumerate(main_src):
            if line.strip().startswith('!'):
                bashcmd = line.strip()[1:]
                main_src[i] = (f"bashcmd = '{bashcmd}'\n"
                            f"process = subprocess.Popen(bashcmd.split(), stdout=subprocess.PIPE)\n"
                            f"output, _ = process.communicate()\n"
                            f"print(output.decode('utf-8'))\n")
        main_src = '\n'.join(main_src)        
        nb_path.with_suffix('.py').write_text(main_src)

        n_lineno = len(str(len(main_src.split('\n'))))
        for i, line in enumerate(main_src.split('\n')):
            print(f'{i:>{n_lineno}}| {line}')
                    
    def upload_all(self):
        with zipfile.ZipFile(f'/tmp/{self.zipname}', 'w', zipfile.ZIP_DEFLATED) as zip_obj:
            zip_obj.write(self.notebook)
            zip_obj.writestr('nb_convert.py', self.get_converter_source())
            if self.dependencies:
                for file in self.dependencies.glob('**/*'):
                    zip_obj.write(file)
        
        cmd = f'gsutil -m cp -r /tmp/{self.zipname} gs://{self.gcs_bucket}/{self.gcs_directory}/{self.zipname}'
        process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        
        return self


class PythonFileOp(dsl.ContainerOp):
    """Represents an op implemented by a PythonFile and dependencies.
    
    Args:
        pyfile: path to PythonFile.
        dependencies: path to dependencies dir.
        gcs_bucket: Google Cloud Storage bucket.
        gcs_directory: GCS dir in gcs_bucket.
        version: tag version for each pythonfile run. If not define,
            version will be set to `%Y-%m-%d_%H:%M:%S` format.
        kwargs: all the arguments for ContainerOp.
        
    Note: 
        - This Op need to be initialized inside pipeline definition or inside a function
            that return an Op for triggering upload source to GCS whenever sources are updated.
        - PythonFile use comments: `# params` after assign statement to set parameters, 
            others would be use as main source.
        - If arguments include list or dict, we can define them in string.
            Eg. ['--a', '"[1,2,3]"',
                 '--b', '\\"{"key1": 1, "key2": "value"}\\"']
        - Remember to save your Notebook before defining pipeline.
    Example: 
        >>> @dsl.pipeline(
        ...     name="test",
        ...     description="test"
        ... )
        ... def pipeline():
        ...     op = PythonFileOp(
        ...         pyfile='path/to/pyfile.py',
        ...         dependencies='path/to/dependencies',
        ...         gcs_bucket='bucket_name',
        ...         gcs_directory='path/to/gsc_directories',
        ...         version='test',
        ...         name='PythonFileOpTest',
        ...         image='path/image:tag',
        ...         arguments=['--a', '"[1,2,3]"',
        ...                    '--b', '5']
        ...     )
    """
    def __init__(self,
                 pyfile:str,
                 dependencies:str,
                 gcs_bucket:str,
                 gcs_directory:str,
                 version:str=None,
                 **kwargs):
        self.pyfile = Path(pyfile.strip())
        self.dependencies = Path(dependencies.strip('/')) if dependencies else None
        self.gcs_bucket = gcs_bucket.strip('/')
        self.gcs_directory = gcs_directory.strip('/')
        self.zipname = ''
        self.version = datetime.now().strftime("%Y-%m-%d_%H:%M:%S") if not version else version

        assert self.pyfile.suffix == '.py'
        if not pyfile:
            raise ValueError("You need to provide a PythonFile.")
            
        if 'name' not in kwargs:
            raise TypeError("You need to provide a name for the operation.")
        elif not kwargs.get('name'):
            raise ValueError("You need to provide a name for the operation.")
        self.zipname = f"{kwargs['name']}_{self.version}.zip"

        if 'image' not in kwargs:
            raise ValueError("You need to provide an image.")
            
        kwargs['command'] = ['sh', '-c']
        kwargs['arguments'] = ' '.join([
            ('echo "SETUP:\n------" && '
             f'gsutil -m cp -r gs://{self.gcs_bucket}/{self.gcs_directory}/{self.zipname} . && '
             f'python3 -m zipfile -e {self.zipname} . && '
             f'echo "\nSOURCE:\n-------" && '
             f'python3 nb_convert.py && '
             f'echo "\nLOG:\n----" && '
             f'python3 src_{self.pyfile.name}'), 
             ' '.join(kwargs.get('arguments', []))]
        )
        self.upload_all()
        super().__init__(**kwargs)
        
    def get_converter_source(self):
        import inspect, os
        converter_src = list(filter('\n'.__ne__, inspect.getsourcelines(self.convert)[0][1:]))
        common = os.path.commonprefix(converter_src)
        ret = [line[len(common):] for line in converter_src]
        for i,line in enumerate(ret):
            ret[i] = line.replace("pyfile_path = Path(f'{pyfile}')", f"pyfile_path = Path(f'{self.pyfile}')")
        return r'{}'.format(''.join(ret))
        
    def convert(self, pyfile):
        import itertools
        from pathlib import Path

        pyfile_path = Path(f'{pyfile}')
        py = pyfile_path.read_text().split('\n')

        main_src = ('import argparse\n'
                    'import ast\n'
                    'parser = argparse.ArgumentParser()\n')

        lvalues = []
        first_params_idx = len(py)
        for line in reversed(py):
            if not line.replace(' ','').endswith('#params'):
                continue
            sharp_idx = line.find('#')
            expresion = line[:sharp_idx]
            lvalue = expresion.split('=')[0].strip()
            lvalues.append(lvalue)
            i = py.index(line)
            if i < first_params_idx:
                first_params_idx = i
                py.remove(line)
        py.insert(first_params_idx, '{}')

        add_arguments = casts = ''
        for arg in reversed(lvalues):
            add_arguments += f"parser.add_argument('--{arg}', type=str, required=True)\n"
            casts += ("try:\n"
                      f"    {arg} = ast.literal_eval(args.{arg})\n"
                      "except:\n"
                      f"    {arg} = args.{arg}\n")

        main_src = '\n'.join(py).format(''.join([main_src, add_arguments, "args = parser.parse_args()\n", casts]))

        pyfile_path.with_name(f'src_{pyfile_path.name}').write_text(main_src)

        n_lineno = len(str(len(main_src.split('\n'))))
        for i, line in enumerate(main_src.split('\n')):
            print(f'{i:>{n_lineno}}| {line}')

    def upload_all(self):
        with zipfile.ZipFile(f'/tmp/{self.zipname}', 'w', zipfile.ZIP_DEFLATED) as zip_obj:
            zip_obj.write(self.pyfile)
            zip_obj.writestr('nb_convert.py', self.get_converter_source())
            if self.dependencies:
                for file in self.dependencies.glob('**/*'):
                    zip_obj.write(file)
        
        cmd = f'gsutil -m cp -r /tmp/{self.zipname} gs://{self.gcs_bucket}/{self.gcs_directory}/{self.zipname}'
        process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        
        return self
    
    



import kfp
from ._client import *
from ._operators import *
from ._component_store import *
from ._components import *
from ._auth import *
from ._pipeline import *

kfp.Client.page_size = 500
kfp.Client.get_pipeline_by_name = get_pipeline_by_name
kfp.Client.get_experiment_by_name = get_experiment_by_name
kfp.Client.get_run_by_name = get_run_by_name
kfp.Client.get_runs_by_experiment = get_runs_by_experiment
kfp.Client.delete_pipeline_by_name = delete_pipeline_by_name
kfp.Client.delete_experiment_by_name = delete_experiment_by_name
kfp.Client.delete_run_by_name = delete_run_by_name
kfp.Client.delete_runs_by_experiment = delete_runs_by_experiment
kfp.Client.upload_pipeline_version = upload_pipeline_version

kfp.compiler.Compiler.__create_pipeline_workflow = kfp.compiler.Compiler._create_pipeline_workflow
kfp.compiler.Compiler._create_pipeline_workflow = create_pipeline_workflow
kfp.dsl.PipelineConf.gcp_sa_name = None
kfp.dsl.PipelineConf.set_gcp_sa = set_gcp_sa


import os
import setuptools
from pathlib import Path

with open(os.path.join(os.getcwd(), 'configs/__VERSION__')) as fp:
    __VERSION__ = str(fp.read())

long_description = Path("README.md").read_text()

setuptools.setup(
    name="kfpxtend",
    version=__VERSION__,
    author="Le Tuan Vu",
    author_email="ltnv24@gmail.com",
    description="High level of kubeflow SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=['kfpxtend'],
    install_requires=['kfp'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)
